package zookeeper

import (
	"context"
	"flag"
	"fmt"
	"strconv"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
)

var log = _log.New("zookeeper")

var confTimeout = 3 * time.Second
var confInterval = time.Duration(0)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "zookeeper.timeout", confTimeout, "[zookeeper] operation timeout")
	fs.DurationVar(&confInterval, "zookeeper.interval", confInterval, "[zookeeper] interval between polls")
}

var Factory = newFactory()

type zookeeperDaemonFactory struct {
	*daemon.TcpDaemonFactory
}

func newFactory() *zookeeperDaemonFactory {
	tcpDaemonFactory := daemon.NewTcpDaemonFactory("zookeeper").(*daemon.TcpDaemonFactory)

	return &zookeeperDaemonFactory{
		TcpDaemonFactory: tcpDaemonFactory,
	}
}

func (df *zookeeperDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	tcpDaemon := df.TcpDaemonFactory.New(paramsI).(*daemon.TcpDaemon)
	tcpDaemon.Timeout = confTimeout
	tcpDaemon.Interval_ = confInterval
	tcpDaemon.Keepalive = false
	tcpDaemon.Logger = log

	return &zookeeperDaemon{
		TcpDaemon: tcpDaemon,
	}
}

type zookeeperDaemon struct {
	*daemon.TcpDaemon
}

func (d *zookeeperDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	data, err := d.Command(ctx, "mntr\n", []string{""})
	if err != nil {
		m.IncrError("zookeeper", "zookeeper")
		return fmt.Errorf("Error collecting zookeeper '%s' metrics: %w", d.ServiceID, err)
	}

	for _, line := range data {
		kv := strings.SplitN(line, "\t", 2)
		if len(kv) != 2 {
			// TODO log
			continue
		}
		stat := strings.Replace(kv[0], "zk_", "", 1)
		valStr := kv[1]
		val, errFloat := strconv.ParseFloat(valStr, 64)

		labels := map[string]string{
			"namespace":    d.Namespace,
			"service":      d.ServiceID,
			"service_type": "zookeeper",
			"service_addr": d.ServiceAddr,
		}

		switch stat {

		case "avg_latency":
			stat = "latency_seconds_avg"
			/* convert to seconds */
			if errFloat == nil {
				val /= 1000
			}

		case "min_latency":
			stat = "latency_seconds_min"
			/* convert to seconds */
			if errFloat == nil {
				val /= 1000
			}

		case "max_latency":
			stat = "latency_seconds_max"
			/* convert to seconds */
			if errFloat == nil {
				val /= 1000
			}

		case "packets_received":
			stat = "packets"
			labels["direction"] = "in"

		case "packets_sent":
			stat = "packets"
			labels["direction"] = "out"

		case "num_alive_connections":
			stat = "alive_connections"

		case "server_state":
			labels["role"] = valStr
			val = float64(time.Now().Unix())
			errFloat = nil

		case "approximate_data_size":
			stat = "approximate_data_size_bytes"

		case "open_file_descriptor_count":
			stat = "open_files"

		case "max_file_descriptor_count":
			stat = "open_files_max"

		case "znode_count":
			stat = "items"
			labels["type"] = "znode"

		case "watch_count":
			stat = "items"
			labels["type"] = "watch"

		case "ephemerals_count":
			stat = "items"
			labels["type"] = "ephemerals"

		case "last_proposal_size":
			stat = "proposal_size"

		case "min_proposal_size":
			stat = "proposal_size_min"

		case "max_proposal_size":
			stat = "proposal_size_max"

		/* ignore useless metrics */
		case "version":
			continue

		/* by default, keep not listed metrics without changing its name and considering it's a float value */
		default:
		}

		if errFloat != nil {
			// TODO log
			continue
		}

		m.Set(&common.Metric{
			Name:   "zookeeper_" + stat,
			Labels: labels,
			Value:  val,
			TTL:    confInterval * 2,
		})
	}

	return nil
}
